/************************************************
** Date :   2011.5.1 
** SWCC (Smoothed Weighting and Configuration Checking) 
** is a Local Search SAT solver
** Author: Shaowei Cai, shaowei_cai@126.com    
**		   School of EECS, Peking University   
**		   Beijing, China                                          
************************************************/


/************************************************
** Date:   2011.9.17 
** SWCCA (Smoothed Weighting and Configuration Checking with Aspiration) 
** Author : Shaowei Cai, shaowei_cai@126.com    
**		   School of EECS, Peking University   
**		   Beijing, China                   
** CC with aspiration: when there are no goodvar (dscore>0 &&
**	     confChange==1); then, if there are greedyvar (dscore
**	     great enough, i.e., bigger than ave_weight), flip the best one,
**	     otherwise, flip the oldest var in a random unsat clause.                  
************************************************/

/************************************************
** Date: 2012.8.7
** Fix a bug when reading redundent clauses
** by Chuan Luo, Peking University
************************************************/

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>

using namespace std;

/* limits on the size of the problem. */
#define MAX_VARS    500000
#define MAX_CLAUSES 2000000

#define pop(stack) stack[--stack ## _fill_pointer]
#define push(item, stack) stack[stack ## _fill_pointer++] = item

#define sigscore	ave_weight   //significant score needed for aspiration

// Define a data structure for a literal in the SAT problem.
struct lit {
		int             clause_num;		//clause num, begin with 0
     	int             var_num;		//variable num, begin with 1
     	int             sense;			//is 1 for true literals, 0 for false literals.
};


/*parameters of the instance*/
int     num_vars;		//var index from 1 to num_vars
int     num_clauses;		//clause index from 0 to num_clauses-1

/* literal arrays */				
lit*	var_lit[MAX_VARS];				//var_lit[i][j] means the j'th literal of var i.
int		var_lit_count[MAX_VARS];			//amount of literals of each var
lit*	clause_lit[MAX_CLAUSES];			//clause_lit[i][j] means the j'th literal of clause i.
int		clause_lit_count[MAX_CLAUSES]; 			// amount of literals in each clause				
			
/* Information about the variables. */
int     dscore[MAX_VARS];				
int		time_stamp[MAX_VARS];
int		conf_change[MAX_VARS];
int*	var_neighbor[MAX_VARS];
int		var_neighbor_count[MAX_VARS];

/* Information about the clauses */			
int     clause_weight[MAX_CLAUSES];		
int     sat_count[MAX_CLAUSES];			
int		sat_var[MAX_CLAUSES];			

//unsat clauses stack
int		unsat_stack[MAX_CLAUSES];		//store the unsat clause number
int		unsat_stack_fill_pointer;
int		index_in_unsat_stack[MAX_CLAUSES];//which position is a clause in the unsat_stack

//variables in unsat clauses
int		unsatvar_stack[MAX_VARS];		
int		unsatvar_stack_fill_pointer;
int		index_in_unsatvar_stack[MAX_VARS];
int		unsat_app_count[MAX_VARS];		//a varible appears in how many unsat clauses

//configuratio changed decreasing variables (dscore>0 and confchange=1)
int		goodvar_stack[MAX_VARS];		
int		goodvar_stack_fill_pointer;
int		already_in_goodvar_stack[MAX_VARS];

/* Information about solution */
int     cur_soln[MAX_VARS];	//the current solution, with 1's for True variables, and 0's for False variables

/* smooth */
int		ave_weight=1;
int		delta_total_weight=0;
int		threshold;
float	p_scale=0.3;//w=w*p+ave*(1-p)
float	q_scale=0.7;//q=1-p 
int		scale_ave;//scale_ave==ave_weight*q_scale

//cutoff steps
int max_tries = 10000000;
int max_flips = 2000000000;


int build_instance(char *filename)
{
	char            line[1024];
	char            tempstr1[10];
	char            tempstr2[10];
	int             cur_lit;
	int             i,j,v,c;
	
	int temp_lit[MAX_VARS];
	
	ifstream infile(filename);
	if(infile==NULL)
		return 0;

	/*** build problem data structures of the instance ***/
	infile.getline(line,1024);
	while (line[0] != 'p')
		infile.getline(line,1024);

	sscanf(line, "%s %s %d %d", tempstr1, tempstr2, &num_vars, &num_clauses);
	
	if(num_vars>=MAX_VARS)
	{
		cout<<"the number of variables exceeds MAX_VARS, please enlarge MAX_VARS."<<endl;
		exit(-1);
	}
	if( num_clauses>=MAX_CLAUSES)
	{
		cout<<"the number of clauses exceeds MAX_CLAUSES, please enlarge MAX_CLAUSES."<<endl;
		exit(-1);
	}
	
	for (v=1; v<=num_vars; ++v)
		var_lit_count[v] = 0;
		
	//Now, read the clauses, one at a time.
	int lit_redundent,clause_redundent;
	for (c = 0; c < num_clauses; ) 
	{
		clause_redundent=0;
        clause_lit_count[c] = 0;
		
		infile>>cur_lit;
		while (cur_lit != 0) { 
		
			lit_redundent=0;
			for(int p=0; p<clause_lit_count[c]; p++)
			{
				if(cur_lit==temp_lit[p]){
					//cout<<"literal "<<cur_lit<<" redundent in clause "<<c<<endl;
					lit_redundent=1;
					break;
				}
				else if(cur_lit==-temp_lit[p]){
					clause_redundent=1;
					break;
				}
			}
			
			if(lit_redundent==0)
			{
		
				temp_lit[clause_lit_count[c]] = cur_lit;
				clause_lit_count[c]++;
			}
		
			infile>>cur_lit;
		}
		
		if(clause_redundent==0)
		{
			clause_lit[c] = new lit[clause_lit_count[c]];
		
			for(i=0; i<clause_lit_count[c]; ++i)
			{
				clause_lit[c][i].clause_num = c;
				clause_lit[c][i].var_num = abs(temp_lit[i]);
				if (temp_lit[i] > 0) clause_lit[c][i].sense = 1;
					else clause_lit[c][i].sense = 0;
			
				var_lit_count[clause_lit[c][i].var_num]++;
			}
			
			c++;
		}
		else 
		{
			num_clauses--;
			clause_lit_count[c] = 0;
		}
	}
	infile.close();
	
	//creat var literal arrays
	for (v=1; v<=num_vars; ++v)
	{
		var_lit[v] = new lit[var_lit_count[v]];
		var_lit_count[v] = 0;	//reset to 0, for build up the array
	}
	//scan all clauses to build up var literal arrays
	for (c = 0; c < num_clauses; ++c) 
	{
		for(i=0; i<clause_lit_count[c]; ++i)
		{
			v = clause_lit[c][i].var_num;
			var_lit[v][var_lit_count[v]] = clause_lit[c][i];
			++var_lit_count[v];
		}
	}
	
	return 1;
		
}

void build_neighbor_relation()
{
	int		neighbor_flag[MAX_VARS];
	int		i,j,count,v,c;

	for(v=1; v<=num_vars; ++v)
	{
		var_neighbor_count[v] = 0;

		for(i=1; i<=num_vars; ++i)
			neighbor_flag[i] = 0;
		neighbor_flag[v] = 1;
		
		for(i=0; i<var_lit_count[v]; ++i)
		{
			c = var_lit[v][i].clause_num;
			for(j=0; j<clause_lit_count[c]; ++j)
			{
				if(neighbor_flag[clause_lit[c][j].var_num]==0)
				{
					var_neighbor_count[v]++;
					neighbor_flag[clause_lit[c][j].var_num] = 1;
				}
			}
		}

		neighbor_flag[v] = 0;
 
		var_neighbor[v] = new int[var_neighbor_count[v]];

		count = 0;
		for(i=1; i<=num_vars; ++i)
		{
			if(neighbor_flag[i]==1)
			{
				var_neighbor[v][count] = i;
				count++;
			}
		}
	}
}


void free_memory()
{
	int i;
	for (i = 0; i < num_clauses; i++) 
		delete[] clause_lit[i];
	
	for(i=1; i<=num_vars; ++i)
	{
		delete[] var_lit[i];
		delete[] var_neighbor[i];
	}
}


inline void unsat(int clause)
{
	index_in_unsat_stack[clause] = unsat_stack_fill_pointer;
	push(clause,unsat_stack);
	
	//update appreance count of each var in unsat clause and update stack of vars in unsat clauses
	int j,v;
	for(j=0; j<clause_lit_count[clause]; ++j)
	{	
		v = clause_lit[clause][j].var_num;
		unsat_app_count[v]++;
		if(unsat_app_count[v]==1)
		{
			index_in_unsatvar_stack[v] = unsatvar_stack_fill_pointer;
			push(v,unsatvar_stack);	
		}
	}
}


inline void sat(int clause)
{
	int index,last_unsat_clause;

	//since the clause is satisfied, its position can be reused to store the last_unsat_clause
	last_unsat_clause = pop(unsat_stack);
	index = index_in_unsat_stack[clause];
	unsat_stack[index] = last_unsat_clause;
	index_in_unsat_stack[last_unsat_clause] = index;
	
	//update appreance count of each var in unsat clause and update stack of vars in unsat clauses
	int j,v,last_unsat_var;
	for(j=0; j<clause_lit_count[clause]; ++j)
	{	
		v = clause_lit[clause][j].var_num;
		unsat_app_count[v]--;
		if(unsat_app_count[v]==0)
		{
			last_unsat_var = pop(unsatvar_stack);
			index = index_in_unsatvar_stack[v];
			unsatvar_stack[index] = last_unsat_var;
			index_in_unsatvar_stack[last_unsat_var] = index;
		}
	}
}

//initiation of the algorithm
void init()
{
	int 		v,c;
	int			i,j;
	int			clause;
	
	//Initialize edge weights
	for (c = 0; c<num_clauses; c++)
		clause_weight[c] = 1;

	//init unsat_stack
	unsat_stack_fill_pointer = 0;
	unsatvar_stack_fill_pointer = 0;

	//build neighbor relation
	build_neighbor_relation();

	//init solution
	for (v = 1; v <= num_vars; v++) {

		if(rand()%2==1) cur_soln[v] = 1;
		else cur_soln[v] = 0;

		time_stamp[v] = 0;
		conf_change[v] = 1;
		unsat_app_count[v] = 0;
	}

	/* figure out sat_count, and init unsat_stack */
	for (c=0; c<num_clauses; ++c) 
	{
		sat_count[c] = 0;
		
		for(j=0; j<clause_lit_count[c]; ++j)
		{
			if (cur_soln[clause_lit[c][j].var_num] == clause_lit[c][j].sense)
			{
				sat_count[c]++;
				sat_var[c] = clause_lit[c][j].var_num;	
			}
		}

		if (sat_count[c] == 0) 
			unsat(c);
	}

	/*figure out var dscore*/
	int lit_count;
	for (v=1; v<=num_vars; v++) 
	{
		dscore[v] = 0;
		
		lit_count = var_lit_count[v];
		
		for(i=0; i<lit_count; ++i)
		{
			c = var_lit[v][i].clause_num;
			if (sat_count[c]==0) dscore[v]++;
			else if (sat_count[c]==1 && var_lit[v][i].sense==cur_soln[v]) dscore[v]--;
		}
	}
		
	//init goodvars stack
	goodvar_stack_fill_pointer = 0;
	for (v=1; v<=num_vars; v++) 
	{
		if(dscore[v]>0)// && conf_change[v]==1)
		{
			already_in_goodvar_stack[v] = 1;
			push(v,goodvar_stack);
			
		}
		else
			already_in_goodvar_stack[v] = 0;
	}
	
	//setting for the virtual var 0
	time_stamp[0]=0;
	
	threshold=300;  //for random 3SAT it's better to set threshold=200+(num_vars+250)/500;
	scale_ave=(threshold+1)*q_scale;//when smoothing, ave_weight=threshold+1.

}

 
//flip a var, and do the neccessary updating
void flip(int flipvar)
{
	int i,j;
	int v,c;
	int index,last_goodvar;
	lit* clause_c;

	int org_flipvar_dscore = dscore[flipvar];
	cur_soln[flipvar] = 1 - cur_soln[flipvar];
	
	//update related clauses and neighbor vars
	for(i=0; i<var_lit_count[flipvar]; ++i)
	{
		c = var_lit[flipvar][i].clause_num;
		clause_c = clause_lit[c];
		if(cur_soln[flipvar] == var_lit[flipvar][i].sense)
		{
			++sat_count[c];
			
			if (sat_count[c] == 2) //sat_count from 1 to 2
				dscore[sat_var[c]] += clause_weight[c];
			else if (sat_count[c] == 1) // sat_count from 0 to 1
			{
				sat_var[c] = flipvar;//record the only true lit's var
				
				for(j=0; j<clause_lit_count[c]; ++j)
					dscore[clause_c[j].var_num] -= clause_weight[c];

				sat(c);
			}
		}
		else // cur_soln[flipvar] != cur_lit.sense
		{
			--sat_count[c];
			if (sat_count[c] == 1) //sat_count from 2 to 1
			{
				for(j=0; j<clause_lit_count[c]; ++j)
				{
					if(clause_c[j].sense == cur_soln[clause_c[j].var_num] )
					{
						dscore[clause_c[j].var_num] -= clause_weight[c];
						sat_var[c] = clause_c[j].var_num;
						break;
					}
				}
			}
			else if (sat_count[c] == 0) //sat_count from 1 to 0
			{
				for(j=0; j<clause_lit_count[c]; ++j)
					dscore[clause_c[j].var_num] += clause_weight[c];
				unsat(c);
			}//end else if
			
		}//end else
	}

	//update information of flipvar
	dscore[flipvar] = -org_flipvar_dscore;
	conf_change[flipvar] = 0;
	
	//remove the vars no longer goodvar in goodvar stack 
	for(index=goodvar_stack_fill_pointer-1; index>=0; index--)
	{
		v = goodvar_stack[index];
		if(dscore[v]<=0)
		{
			goodvar_stack[index] = pop(goodvar_stack);
			already_in_goodvar_stack[v] = 0;
		}	
	}

	//update all flipvar's neighbor's conf_change to be 1, add goodvar
	for(i=0; i<var_neighbor_count[flipvar]; ++i)
	{
		v = var_neighbor[flipvar][i];
		conf_change[v] = 1;
		
		if(dscore[v]>0 && already_in_goodvar_stack[v] ==0)
		{
			push(v,goodvar_stack);
			already_in_goodvar_stack[v] = 1;
		}
	}
 
}


/*the following functions are non-algorithmic*/

void print_solution()
{
     cout<<"v ";
     for (int i=1; i<=num_vars; i++) {
		if(cur_soln[i]==0) cout<<"-";
		cout<<i;
		if(i%10==0) cout<<endl<<"v ";
		else	cout<<' ';
     }
     cout<<"0"<<endl;
}

int verify_sol()
{
	int c,j,flag;
	for (c = 0; c<num_clauses; ++c) 
	{
		flag = 0;
		for(j=0; j<clause_lit_count[c]; ++j)
			if (cur_soln[clause_lit[c][j].var_num] == clause_lit[c][j].sense) {flag = 1; break;}

		if(flag ==0){//output the clause unsatisfied by the solution
			cout<<"clause "<<c<<" is not satisfied"<<endl;
			for(j=0; j<clause_lit_count[c]; ++j)
			{
				if(clause_lit[c][j].sense==0)cout<<"-";
				cout<<clause_lit[c][j].var_num<<" ";
			}
			cout<<endl;
			for(j=0; j<clause_lit_count[c]; ++j)
				cout<<cur_soln[clause_lit[c][j].var_num]<<" ";
			cout<<endl;
			return 0;
		}
	}
	return 1;
}

